    %% Example: Bus aggregation methodology
% Based on the paper: An improved Bus Aggregation Technique for Generating Network Equivalents, Di Shi et.al., 2012.
% Author: Andr Ortner
% Date: 22.10.2013

%% Definitions
%{
bus = node
branch = line
cluster = group of nodes
zone = assigned group of countries
%}

clear all
clc
%% load country list of numbers, abbreviation and names
load 'countrylist.mat'

%%
node_injections = xlsread('node_injections.xlsx','Tabelle1','A2:AJX10000');
num_scenarios = length(node_injections(1,:));
%%  -----  Power Flow Data  -----%%
mpc.bus = xlsread('nodes.xls','nodes','A2:K10000');
num_nodes = length(mpc.bus(:,1));

mpc.bus(:,12) = mean(node_injections')'; %generation of uniformly distribution generator outputs
   
%% branch data (also including transformers)
mpc.branch = xlsread('lines.xls','lines','A2:S10000');
num_lines = length(mpc.branch(:,1));

run('zone_numbering.m') 

%% -----  PTDF calculation  -----%%
% Build the incidence matrix 
% ATTENTION: brackets are necessary
% ATTENTION: order of commands in crucial
C_Inc = (kron(ones(1,length(mpc.bus(:,1))),mpc.branch(:,18)) == ones(length(mpc.branch(:,1)),1) * [1:length(mpc.bus(:,1))]) + ...  % Positive indices 
        (-1).*(kron(ones(1,length(mpc.bus(:,1))),mpc.branch(:,19)) == ones(length(mpc.branch(:,1)),1) * [1:length(mpc.bus(:,1))]); % Negative indices

% Build the reduced incidence matrix (w/o slack node)
C_IncWoS = C_Inc; 
C_IncWoS(:,mpc.bus(:,8)==3) = [];
% Build the reduced node (bus) matrix (w/o slack node)
mpc.busWoS = mpc.bus;
mpc.busWoS(mpc.busWoS(:,8)==3,:) = [];

B_flow = diag(mpc.branch(:,12).^(-1)) * C_IncWoS;
B_bus = C_IncWoS.' * B_flow;

[U, S, V]= svd(B_bus);
s = diag(S); 
k = sum(s> 1e-9); % simple thresholding based decision
B_bus_inv = (U(:, 1: k)* diag(1./ s(1: k))* V(:, 1: k)')';

PTDF = B_flow * B_bus_inv;
 
% Results for the power flows of the original grid
P_flow_lines = PTDF * mpc.busWoS(:,12);
%%
P_flow = zeros(num_lines,num_scenarios);

for i = 1:num_scenarios
    P_flow(:,i) = PTDF * node_injections(2:end,i);
end

%%
exceed_counter = zeros(num_lines,1);
max_exceed = zeros(num_lines,1);

for i = 1:num_lines
   exceed_counter(i) = length(P_flow(i,abs(P_flow(i,:))>mpc.branch(i,13)));  
   max_exceed(i) = max(P_flow(i,:))-mpc.branch(i,13);
end

%%
 disp('cluster process');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));

number_zones = max(mpc.bus(:,6));
num_nodes_in_zone = zeros(number_zones,1);
num_clusters_in_zone = zeros(number_zones,1);
target_num_clusters_in_zone = zeros(number_zones,1);
num_lines_aggregate_zone = zeros(number_zones,1);

clusterintensity = 0.2;
number_nodes = length(mpc.bus(:,1));
num_lines = length(mpc.branch(:,1));
mpc.bus(:,7) = 0;
g=0;

% assign cluster ID's based on clustering condition, clustering process gets repeated while enheightening the condition treshold
% until desired zone size is reached

for j=1:number_zones
    
    
    zone_nodes_IND = find(mpc.bus(:,6) ==j); %
    zone_nodes = mpc.bus(zone_nodes_IND,:); %selection of only nodes of the current zone
    num_nodes_in_zone(j,1) = length(mpc.bus(mpc.bus(:,6) ==j)); 
    
    target_num_clusters_in_zone(j,1) = clusterintensity*num_nodes_in_zone(j,1); % number of clusters in zone is initally equal to the number of nodes 
    num_lines_aggregate_zone(j,1) = num_nodes_in_zone(j,1) - target_num_clusters_in_zone(j,1);
    
    zone_branches_IND = find(mpc.branch(:,21)==j & mpc.branch(:,22)==j);
    zone_branches = mpc.branch(zone_branches_IND,:); % selection of only branches IN the current zone
    num_branches_zone = length(zone_branches(:,1));
    zone_exceed_counter = exceed_counter(zone_branches_IND,:);
    zone_max_exceed = max_exceed(zone_branches_IND,:);
    
    if ~isempty(zone_nodes(:,1))
    
        for i=1:num_lines_aggregate_zone(j,1)
          max_number_exceed_IND = find(zone_exceed_counter == max(zone_exceed_counter));
          zone_max_exceed_IND = find(zone_max_exceed(max_number_exceed_IND) == max(zone_max_exceed(max_number_exceed_IND)));
          line_remove_IND = max_number_exceed_IND(zone_max_exceed_IND); % index of line in matrix zone_branches
          
                %create clustersID's  
                    from_IND = find(mpc.bus(:,11) == zone_branches(line_remove_IND,18)); %find the from_node of a branch in the zone nodes
                    
                    to_IND = find(mpc.bus(:,11) == zone_branches(line_remove_IND,19));  %find the to_node of a branch in the zone nodes
                    
                    if  mpc.bus(from_IND,7) ~= 0 && mpc.bus(to_IND,7) ~= 0 % if both nodes from a branch already have a cluster number, the to_node gets the number of the from_node
                   
                    replace_cluster = mpc.bus(to_IND,7);
                    for l=1:number_nodes
                        if mpc.bus(l,7) == replace_cluster;
                             
                            mpc.bus(l,7) = mpc.bus(from_IND,7);
                        end
                    end
                    
                elseif mpc.bus(from_IND,7) ~= 0 % if only the from_node already has a cluster number, the to_node gets the number of the from_node
                    mpc.bus(to_IND,7)= mpc.bus(from_IND,7);
                elseif mpc.bus(to_IND,7) ~= 0 % if only the to_node already has a cluster number, the from_node gets the number of the to_node
                    mpc.bus(from_IND,7) = mpc.bus(to_IND,7);
                else % if none of the nodes of a branch has a cluster number, the cluster counter is going up 1, and both from_node and to_node get this cluster number assigned
                        
                    g=g+1;
                    mpc.bus(from_IND,7) =g;
                    mpc.bus(to_IND,7) =g;
                    end  
                zone_branches(line_remove_IND,:) = [];
                zone_exceed_counter(line_remove_IND,:) = [];
                zone_max_exceed(line_remove_IND,:) = [];
              end   
            end  
        
        %number all unnumbered nodes in the zone
        for b=1:number_nodes
            if mpc.bus(b,6) == j
            if mpc.bus(b,7) == 0
              g = g+1;
              mpc.bus(b,7) = g;         
            end         
            end 
         end
end

%renumber from 1 to num_clusters,
replace_cluster = zeros(number_nodes,1);
num_clusters = length(unique(mpc.bus(:,7)));
for i=1:num_clusters 
rep_clus = find(mpc.bus(:,7) == max(mpc.bus(:,7)));
replace_cluster(rep_clus,1) = i;
mpc.bus(rep_clus,7)=0;
end

mpc.bus(:,7) = replace_cluster(:,1);

% create matrix describing which which cluster belongs to which node
mpc.clusterNodeSumID = zeros(num_clusters,number_nodes); %matrix containing the indices of node injections in mpc.bus of zone i of the reduced grid
for i=1:num_clusters, 
    mpc.clusterNodeSumID(i,:) = ismember(mpc.bus(:,7),i)'; 
end;


%% ----- Build reduced grid equivalent ----- %%

 disp('build reduced grid equivalent');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));


% Define orientations AND connections of branches from the equivalent grid
% First approach (maybe inefficient) try all combinations
% => for i=1:number of Zones: (for all zones in Z ~= Z(i)) ...
% => for all x in Z(i): if one y from Z(j) exists & (x,y) element of mcp.branch -> (Z(i),Z(j)) is element of mcp.branchR 
% NOTE: A more efficient approach would be to restrict possible zone combinations to neighbour zones (could be done via introducing area/zone combinations in mcp.bus)
% NOTE: The most efficient way would be to predefine the connection matrix
mpc.branchR = [];
mpc.clusterBranchSumID = [];

k = 0; %running variable that increases with the development of mpc.branchR 
for i=1:num_clusters, %from cluster
    for j=1:num_clusters, %to cluster
        if(j > i)
        clusterID_j = find(mpc.bus(:,7) == j);
        clusterID_i = find(mpc.bus(:,7) == i);
        [pairs_i, pairs_j] = meshgrid(mpc.bus(clusterID_i,11),mpc.bus(clusterID_j,11));
        pairs_ij = [pairs_i(:) pairs_j(:)];
        pairs_ji = [pairs_j(:) pairs_i(:)];
        %other solution -> more complicated because intra-zonal flows has to be removed
        %pairs_ij = combnk([mpc.bus(clusterID_i,1); mpc.bus(clusterID_j,1)],2); %all possible line-pair combinations (no repetition) containing a node of cluster i and cluster j
        %pairs_ij(~ismember(pairs_ij(:,1),mpc.bus(clusterID_i,1)),:) = []; %remove all combinations that are intra-zonal flows (= do not start in cluster 1)
        pairs_all = mpc.branch(:,18:19); %all existing connections of the original grid
        [pairs_ijR_1, pairs_allID_ij, pairs_ijID] = intersect( pairs_all, pairs_ij, 'rows');
        
 % PROBLEM THAT NEG. ORIENTED PAIR GET NOT INTO THE LIST
        [pairs_jiR_2, pairs_allID_ji, pairs_jiID] = intersect( pairs_all, pairs_ji, 'rows'); %lines in different orientation than in between clusters should also be counted
        pairs_ijR = [pairs_ijR_1; pairs_jiR_2];
        pairs_allID = [pairs_allID_ij; pairs_allID_ji];
        
            if( ~isempty(pairs_ijR)), %Do mcp.branch contain a line between Z(i) and Z(j)?
                k=k+1; 
                BranchSum = ( ismember(mpc.branch(:,18), mpc.bus(clusterID_i,11)).' & ismember(1:length(mpc.branch(:,18)), pairs_allID) ).* 1 + ...
                            ( ismember(mpc.branch(:,18), mpc.bus(clusterID_j,11)).' & ismember(1:length(mpc.branch(:,18)), pairs_allID) ).*(-1);
                mpc.clusterBranchSumID(k,:) = BranchSum; %all indices of
                if i<j, %branch orientation of reduced grid is chosen in a way that orientation of flows go from clusters with smallest ID to the ones with higher IDs 
                mpc.branchR = [mpc.branchR;[0 i   j	0	0	1	0	0	0	0	0	0	0]]; %initial values of reduced branch matrix without defining the orientation
                else mpc.branchR = [mpc.branchR;[0  j	i	0	0	1	0	0	0	0	0	0   0]]; 
                end
            end
        end     
    end
end

mpc.branchR(:,1) = unique(1:length(mpc.branchR(:,1)));
num_linesR = length(mpc.branchR(:,1));

% create reduced bus matrix (clusters), based on the combined properties of
% the old bus matrix
mpc.busR = zeros(num_clusters,12);
 %first colum of the matrix is numbered

for i =1:num_clusters
    node_IND = find(mpc.bus(:,7) == i); % find cluster numbers in node matrix
    mpc.busR(i,1) = mpc.bus(node_IND(1),7); % cluster number
    mpc.busR(i,2) = mpc.bus(node_IND(1),6); % country
    mpc.busR(i,9) = mean(mpc.bus(node_IND,9)); %X coordinate
    mpc.busR(i,10) = mean(mpc.bus(node_IND,10)); %Y coordinate
    mpc.busR(i,12) = sum(mpc.bus(node_IND,12)); %injection
end

mpc.busR(:,8) = 2; %slack node
mpc.busR(1,8) = 3; %first zone (equivalent bus) is defined as slack node

% fill in reduced branch matrix

for i = 1:num_lines
mpc.branch(i,23) = mpc.bus(mpc.branch(i,18),7);
mpc.branch(i,24) = mpc.bus(mpc.branch(i,19),7);
end


%%
 disp('calculate reduced PTDF');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));
 
node_IND = find(mpc.bus(:,8)==2);

P_flow_clusters = mpc.clusterBranchSumID * P_flow_lines; %%matrix containing the sum of power flows resulting from the original grid in between the zones according to zone definition in mcp.branchR
% NOTE:
% - The dimension of the matrices is important in order to split the matrix in 2 products, whereas one into a quadratic size that can be inverted.

% ----- Calculate the PDTF of the reduced grid ----- %%
gamma_R = mpc.clusterNodeSumID(2:end,node_IND); %maps the indices of power injections from the original grid nodes to clusters
% The PTDF_R are being calculated by setting the intra-clusteral clusters equal to the branch flows of the reduced grid

% take out zeros in matrix, so no singular matrix will occur
matrix = gamma_R * mpc.bus(node_IND,12);
for i=1:length(matrix(:,1))
    if matrix(i,1) == 0
        matrix(i,1) = 0.1;
    end
end


PTDF_R = mpc.clusterBranchSumID * PTDF * diag(mpc.bus(node_IND,12)) * gamma_R.' * (diag(matrix))^(-1); %Power flows are collected just for the reduced network (without slack)


%% ----- Calculate the reactances of the reduced grid ----- %%

% Build the incidence matrix of the reduced grid
% ATTENTION: brackets are necessary
% ATTENTION: order of commands is crucial
C_IncR = (kron(ones(1,length(mpc.busR(:,1))),mpc.branchR(:,2)) == ones(length(mpc.branchR(:,1)),1) * [1:length(mpc.busR(:,1))]) + ...  % Positive indices 
         (-1).*(kron(ones(1,length(mpc.busR(:,1))),mpc.branchR(:,3)) == ones(length(mpc.branchR(:,1)),1) * [1:length(mpc.busR(:,1))]); % Negative indices

% Build the reduced incidence matrix (w/o slack node)
C_IncRWoS = C_IncR; 
C_IncRWoS(:,mpc.busR(:,8)==3) = [];
% Build the reduced node (bus) matrix (w/o slack node)
mpc.busRWoS = mpc.busR;
mpc.busRWoS(mpc.busRWoS(:,8)==3,:) = [];

LR = length(mpc.branchR(:,1));
NR = length(mpc.busRWoS(:,1));

%Arbitrary matrix 
Lambda = zeros(LR*NR,LR);
for i=1:NR,
Lambda(1+(i-1)*LR:LR+(i-1)*LR,:) = (PTDF_R * C_IncRWoS.' - eye(LR)) * diag(C_IncRWoS(:,i));
end

%Calculate average angle difference between zones
%NOTE: In a further step the NODE angles should be averaged in each zone and from that the difference should be taken phi* = phi1* - phi2*
% - This calculation should be the result of a DC power flow calculation with the original grid
% - But what about the varying averages of the angles over time? Reduce quadratic error of time sequence?
theta_diff = P_flow_lines .* mpc.branch(:,12);
mpc.clusterBranchSumID_test = mpc.clusterBranchSumID;

theta_diff_cluster = abs(mpc.clusterBranchSumID * theta_diff./sum(abs(mpc.clusterBranchSumID)')');

%{
for i = 1:num_linesR
    if theta_diff_cluster(i,1) <= 0
    theta_diff_cluster(i,1) = theta_diff_cluster(i,1)*-1;
    cluster_from = mpc.branchR(i,2);
    cluster_to = mpc.branchR(i,3);
    %mpc.branchR(i,2)= cluster_to;
    %mpc.branchR(i,3) = cluster_from; 
    end
end
%}


%Calculation of the reactances
%NOTE: There is no unique solution to derive from PDTF data reactances -> they should be physically correct, therefore angle limitations from the original grid are used as benchmark

Lambda_ext = [eye(length(theta_diff_cluster)); Lambda]; %Additional side constraints that angle difference of reduced grid should be in the range of the intra-zonal average angle differences of the orignial gird
b_ext = [abs(P_flow_clusters) ./ theta_diff_cluster; zeros(length(Lambda),1)]; %ride-hand side of the reactances equation
%xR_inv = (Lambda_ext.' * Lambda_ext)^(-1) * Lambda_ext.' * b_ext; %the solution of the equation Lambda_ext * (1/x_R) = b_ext
xR_inv = Lambda_ext \ b_ext; %straight solution of the equation without former reshaping of the equation

mpc.branchR(:,5) = abs(1./ xR_inv); %the reactances of the reduced grid are written into the reduced grid branch matrix

for i = 1:num_linesR
    line_IND11 = find(mpc.branch(:,23) == mpc.branchR(i,2));
    line_IND12 = find(mpc.branch(:,24) == mpc.branchR(i,3));
    line_IND1 = intersect(line_IND11,line_IND12);
    
    line_IND21 = find(mpc.branch(:,24) == mpc.branchR(i,2));
    line_IND22 = find(mpc.branch(:,23) == mpc.branchR(i,3));
    line_IND2 = intersect(line_IND21,line_IND22);
    line_IND = [line_IND1; line_IND2];

     mpc.branchR(i,4) = 1/sum(1./mpc.branch(line_IND,11));
     mpc.branchR(i,6) = sum(mpc.branch(line_IND,13));
end
%}
%% create matrix describing which cluster belongs to which zone
mpc.clusterZoneID=zeros(num_clusters,1);

for i=1:number_nodes
    mpc.clusterZoneID(mpc.bus(i,7),1) = mpc.bus(i,6);
end


number_scenarios = 959;

%preallocation of memory
Error_abs = zeros(number_scenarios,length(PTDF_R(:,1)));
Error_rel = zeros(number_scenarios,num_linesR);

num_linesR = length(mpc.branchR(:,1));
%%
 disp('calculate absolute error');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));
node_IND = find(mpc.bus(:,8)==2);

for i=1:number_scenarios
P_flow_clusters_actual = mpc.clusterBranchSumID * PTDF * node_injections(node_IND,i+1);
P_flow_clusters_reduced = PTDF_R * gamma_R * node_injections(node_IND,i+1);
      
Error_abs(i,:) = abs(P_flow_clusters_reduced-P_flow_clusters_actual);  % absolute error  
     
end

%% measure error relative to total NTC values
 disp('calculate relative error');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));


NTC_flows = (abs(mpc.clusterBranchSumID)*abs(mpc.branch(:,13)))'; % the total capacity between two nodes

for i=1:num_linesR
Error_rel(:,i)=Error_abs(:,i) /NTC_flows(1,i); %  % relative error compared to total capacity between two nodes
end

Error_rel_av = mean(Error_rel)'; % average error of each flow 
%% create matrix with per zone the list of flows in it

 disp('calculate error per zone');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));

zone_counter=zeros(number_zones,1);
flowZoneID = [];
for i=1:num_linesR
    if mpc.clusterZoneID(mpc.branchR(i,2))~= mpc.clusterZoneID(mpc.branchR(i,3))
    zoneID = mpc.clusterZoneID(mpc.branchR(i,2));
    zone_counter(zoneID,1) = zone_counter(zoneID,1) +1;
    flowZoneID(zone_counter(zoneID,1),zoneID)=i;
    zoneID = mpc.clusterZoneID(mpc.branchR(i,3));
    zone_counter(zoneID,1) = zone_counter(zoneID,1) +1;
    flowZoneID(zone_counter(zoneID,1),zoneID)=i;
    end
end
%%
%create matrix with per zone the list of flowerrors in it
for i=1:number_zones
    for j=1:length(flowZoneID(:,1))
        if flowZoneID(j,i) ~= 0
        Error_per_Zone(j,i) = Error_rel_av(flowZoneID(j,i),1);
        end
    end
end

% take average from errors in flows
average_Error_per_Zone = zeros(1,number_zones);
columnlength = zeros(1,number_zones);

for i=1:number_zones
    if size(find(Error_per_Zone(:,i)~=0),1)~=0
    columnlength(1,i)=max(find(Error_per_Zone(:,i)~=0));
    end
    if columnlength(1,i)~=0 
        average_Error_per_Zone(1,i) = sum(Error_per_Zone(:,i))./columnlength(1,i);
    end
        
end


%% write excel file with the results

 disp('finished');
 c = clock;
 disp(datestr(datenum(c(1),c(2),c(3),c(4),c(5),c(6))));
