%% Example: Bus aggregation methodology
% Based on the paper: An improved Bus Aggregation Technique for Generating Network Equivalents, Di Shi et.al., 2012.
% Author: Andr Ortner
% Date: 22.10.2013

clear all
clc

%% -----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% -----  Power Flow Data  -----%%
%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	cluster	Vmax	Vmin
%    (1)     (2)    (3) (4) (5) (6) (7)     (8) (9) (10)    (11)    (12)    (13)
mpc.bus = [
1	3	-200	0	0	0	1	1.06	0	380	6	1.06	0.94;...
2	2	-100	0	0	0	1	1.06	0	380	6	1.06	0.94;...
3	2	-200	0	0	0	1	1.06	0	380	5	1.06	0.94;...
4	2	1000	0	0	0	1	1.06	0	380	6	1.06	0.94;...
5	2	900     0	0	0	1	1.06	0	380	1	1.06	0.94;...
6	2	-50     0	0	0	1	1.06	0	380	1	1.06	0.94;...
7	2	-50     0	0	0	1	1.06	0	380	1	1.06	0.94;...
8	2	-100	0	0	0	1	1.06	0	380	2	1.06	0.94;...
9	2	1000	0	0	0	1	1.06	0	380	3	1.06	0.94;...
10	2	-500	0	0	0	1	1.06	0	380	4	1.06	0.94;...
11	2	-700	0	0	0	1	1.06	0	380	4	1.06	0.94;...
12	2	-1000	0	0	0	1	1.06	0	380	4	1.06	0.94;...
];

%% branch data (also including transformers)
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
4	2	0	0.25784847	0	0	0	0	0	0	1	-360	360;...
11	10	0	0.5069097	0	0	0	0	0	0	1	-360	360;...
5	6	0	0.496799952	0	0	0	0	0	0	1	-360	360;...
2	1	0	0.395166123	0	0	0	0	0	0	1	-360	360;...
10	12	0	1.067796972	0	0	0	0	0	0	1	-360	360;...
6	7	0	0.744602232	0	0	0	0	0	0	1	-360	360;...
2	3	0	0.541057077	0	0	0	0	0	0	1	-360	360;...
10	9	0	1.60581069	0	0	0	0	0	0	1	-360	360;...
6	8	0	1.585496352	0	0	0	0	0	0	1	-360	360;...
4	5	0	1.932910122	0	0	0	0	0	0	1	-360	360;...
5	11	0	3.856535958	0	0	0	0	0	0	1	-360	360;...
11	4	0	5.874174999	0	0	0	0	0	0	1	-360	360;...
];
      

%% -----  PTDF calculation  -----%%
% Build the incidence matrix 
% ATTENTION: brackets are necessary
% ATTENTION: order of commands in crucial
C_Inc = (kron(ones(1,length(mpc.bus(:,1))),mpc.branch(:,1)) == ones(length(mpc.branch(:,1)),1) * [1:length(mpc.bus(:,1))]) + ...  % Positive indices 
        (-1).*(kron(ones(1,length(mpc.bus(:,1))),mpc.branch(:,2)) == ones(length(mpc.branch(:,1)),1) * [1:length(mpc.bus(:,1))]); % Negative indices

% Build the reduced incidence matrix (w/o slack node)
C_IncWoS = C_Inc; 
C_IncWoS(:,mpc.bus(:,2)==3) = [];
% Build the reduced node (bus) matrix (w/o slack node)
mpc.busWoS = mpc.bus;
mpc.busWoS(mpc.busWoS(:,2)==3,:) = [];

B_flow = diag(mpc.branch(:,4).^(-1)) * C_IncWoS;
B_bus = C_IncWoS.' * B_flow;
PTDF = B_flow * B_bus^(-1);

% Results for the power flows of the original grid
P_flow_lines = PTDF * mpc.busWoS(:,3);


%% ----- Build reduced grid equivalent according to zone definition in mcp.bus ----- %%
% -- Initialize reduced bus matrix (suffix 'R') --
num_zones = length(unique(mpc.bus(:,11)));  %should be of the same form than mpc.bus
mpc.busR = zeros(num_zones,length(mpc.bus(1,:)));
mpc.busR(:,1) = unique(mpc.bus(:,11)); %first colum of the matrix is numbered

%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.busR(:,2:end) = kron(ones(num_zones,1),...
[2   0   0   0   0   1   1.06    0   380 1   1.06    0.94]); %initial values of reduced bus matrix without bus nummer (=colum 1 of bus matrix)
% NOTE: 
% - all nodes are in Zone 1
% - the node type is defined to 2 (PU-node)
% - all reference voltages are set to 380kV
mpc.busR(1,2) = 3; %first zone (equivalent bus) is defined as slack node
mpc.zoneNodeSumID = zeros(num_zones,length(mpc.bus(:,1))); %matrix containing the indices of node injections in mpc.bus of zone i of the reduced grid
for i=1:num_zones, 
    mpc.zoneNodeSumID(i,:) = ismember(mpc.bus(:,11),i)'; 
end;
mpc.busR(:,3) = mpc.zoneNodeSumID * mpc.bus(:,3); 
% NOTE: 
% - "ismember" might be replaced by "ismemberc", which is much faster
% - the (time) row of the values of the power inflows has to be filled in

% -- Initialize reduced branch matrix (suffix 'R') -- 
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branchR = []; %reduced branch matrix
mpc.zoneBranchSumID = []; %matrix containing the indices of lines in mpc.branch, whose flows go from one zone to a certain other zone in the reduced grid 

% Define orientations AND connections of branches from the equivalent grid
% First approach (maybe inefficient) try all combinations
% => for i=1:number of Zones: (for all zones in Z ~= Z(i)) ...
% => for all x in Z(i): if one y from Z(j) exists & (x,y) element of mcp.branch -> (Z(i),Z(j)) is element of mcp.branchR 
% NOTE: A more efficient approach would be to restrict possible zone combinations to neighbour zones (could be done via introducing area/zone combinations in mcp.bus)
% NOTE: The most efficient way would be to predefine the connection matrix

k = 0; %running variable that increases with the development of mpc.branchR 
for i=1:num_zones, %from zone
    for j=1:num_zones, %to zone
        if(j > i)
        zoneID_j = find(mpc.bus(:,11) == j);
        zoneID_i = find(mpc.bus(:,11) == i);
        [pairs_i, pairs_j] = meshgrid(mpc.bus(zoneID_i,1),mpc.bus(zoneID_j,1));
        pairs_ij = [pairs_i(:) pairs_j(:)];
        pairs_ji = [pairs_j(:) pairs_i(:)];
        %other solution -> more complicated because intra-zonal flows has to be removed
        %pairs_ij = combnk([mpc.bus(zoneID_i,1); mpc.bus(zoneID_j,1)],2); %all possible line-pair combinations (no repetition) containing a node of Zone i and Zone j
        %pairs_ij(~ismember(pairs_ij(:,1),mpc.bus(zoneID_i,1)),:) = []; %remove all combinations that are intra-zonal flows (= do not start in zone 1)
        
        pairs_all = mpc.branch(:,1:2); %all existing connections of the original grid
        [pairs_ijR_1, pairs_allID_ij, pairs_ijID] = intersect( pairs_all, pairs_ij, 'rows');
        
 % PROBLEM THAT NEG. ORIENTED PAIR GET NOT INTO THE LIST
        [pairs_jiR_2, pairs_allID_ji, pairs_jiID] = intersect( pairs_all, pairs_ji, 'rows'); %lines in different orientation than in between zones should also be counted
        pairs_ijR = [pairs_ijR_1; pairs_jiR_2];
        pairs_allID = [pairs_allID_ij; pairs_allID_ji];
        
            if( ~isempty(pairs_ijR)), %Do mcp.branch contain a line between Z(i) and Z(j)?
                k=k+1; 
                BranchSum = ( ismember(mpc.branch(:,1), mpc.bus(zoneID_i,1)).' & ismember(1:length(mpc.branch(:,1)), pairs_allID) ).* 1 + ...
                            ( ismember(mpc.branch(:,1), mpc.bus(zoneID_j,1)).' & ismember(1:length(mpc.branch(:,1)), pairs_allID) ).*(-1);
                mpc.zoneBranchSumID(k,:) = BranchSum; %all indices of
                if i<j, %branch orientation of reduced grid is chosen in a way that orientation of flows go from zones with smallest ID to the ones with higher IDs 
                mpc.branchR = [mpc.branchR;[i	j	0	0	0	0	0	0	0	0	1	-360	360]]; %initial values of reduced branch matrix without defining the orientation
                else mpc.branchR = [mpc.branchR;[j	i	0	0	0	0	0	0	0	0	1	-360	360]]; 
                end
            end
        end     
    end
end
% NOTE: The values of the equivalent reactances has to be filled in the matrix later 


%% ----- Calculate the PDTF of the reduced grid ----- %%
P_flow_zones = mpc.zoneBranchSumID * P_flow_lines; %%matrix containing the sum of power flows resulting from the original grid in between the zones according to zone definition in mcp.branchR
% NOTE:
% - The dimension of the matrices is important in order to split the matrix in 2 products, whereas one into a quadratic size that can be inverted.

gamma_R = mpc.zoneNodeSumID(2:end,2:end); %maps the indices of power injections from the original grid nodes to zones
% The PTDF_R are being calculated by setting the intra-zonal zones equal to the branch flows of the reduced grid
PTDF_R = mpc.zoneBranchSumID * PTDF * diag(mpc.bus(2:end,3)) * gamma_R.' * (diag(gamma_R * mpc.bus(2:end,3)))^(-1); %Power flows are collected just for the reduced network (without slack)

%% ----- Calculate the reactances of the reduced grid ----- %%
% Build the incidence matrix of the reduced grid
% ATTENTION: brackets are necessary
% ATTENTION: order of commands is crucial
C_IncR = (kron(ones(1,length(mpc.busR(:,1))),mpc.branchR(:,1)) == ones(length(mpc.branchR(:,1)),1) * [1:length(mpc.busR(:,1))]) + ...  % Positive indices 
         (-1).*(kron(ones(1,length(mpc.busR(:,1))),mpc.branchR(:,2)) == ones(length(mpc.branchR(:,1)),1) * [1:length(mpc.busR(:,1))]); % Negative indices

% Build the reduced incidence matrix (w/o slack node)
C_IncRWoS = C_IncR; 
C_IncRWoS(:,mpc.busR(:,2)==3) = [];
% Build the reduced node (bus) matrix (w/o slack node)
mpc.busRWoS = mpc.busR;
mpc.busRWoS(mpc.busRWoS(:,2)==3,:) = [];

LR = length(mpc.branchR(:,1));
NR = length(mpc.busRWoS(:,1));

%Arbitrary matrix 
Lambda = zeros(LR*NR,LR);
for i=1:NR,
Lambda(1+(i-1)*LR:LR+(i-1)*LR,:) = (PTDF_R * C_IncRWoS.' - eye(LR)) * diag(C_IncRWoS(:,i));
end

%Calculate average angle difference between zones
%NOTE: In a further step the NODE angles should be averaged in each zone and from that the difference should be taken phi* = phi1* - phi2*
% - This calculation should be the result of a DC power flow calculation with the original grid
% - But what about the varying averages of the angles over time? Reduce quadratic error of time sequence?
theta_diff = P_flow_lines .* mpc.branch(:,4);
theta_diff_zone = mpc.zoneBranchSumID * theta_diff ./ sum(mpc.zoneBranchSumID,2);

%Calculation of the reactances
%NOTE: There is no unique solution to derive from PDTF data reactances -> they should be physically correct, therefore angle limitations from the original grid are used as benchmark

Lambda_ext = [eye(length(theta_diff_zone)); Lambda]; %Additional side constraints that angle difference of reduced grid should be in the range of the intra-zonal average angle differences of the orignial gird
b_ext = [P_flow_zones ./ theta_diff_zone; zeros(length(Lambda),1)]; %ride-hand side of the reactances equation
%xR_inv = (Lambda_ext.' * Lambda_ext)^(-1) * Lambda_ext.' * b_ext; %the solution of the equation Lambda_ext * (1/x_R) = b_ext
xR_inv = Lambda_ext \ b_ext; %straight solution of the equation without former reshaping of the equation


%% alternative solution:
%set up of an minimization problem with maximum/minimum barriers for the voltage angles (and as a side condition also for the reactances) 
%obj_fct = @(xR_inv) abs(sum(P_flow_zones) - (theta_diff_zone' * xR_inv));
%xR_max = 10000 * ones(length(P_flow_zones),1);
%xR_min = 0.1 * ones(length(P_flow_zones),1);
%[xR_inv,P_flow_diff] = fmincon(obj_fct,zeros(length(P_flow_zones),1),... % Objective function and initial values
%                       (-1).*eye(length(theta_diff_zone)), (-1).*(P_flow_zones./theta_diff_zone), ... % -A*x <= -b -> A*x >= b
%                       Lambda, zeros(length(Lambda),1)); % Aeq * x = beq
%                       1./xR_max, 1./xR_min); % lower / upper angle limits
%NOTE:
% - there is no solution for Lambda = 0; (an error remains)

mpc.branchR(:,4) = 1./ xR_inv; %the reactances of the reduced grid are written into the reduced grid branch matrix


%% ----- Error evaluation under changed operating conditions ----- %%
% For this evaluation the PDTF_R is used that exactly match the power flows of the original grid during the initial operating point
% P_inj_test = mpc.bus(2:end,3); % Original point of operation on that the reduced grid has been calibrated

rng(0,'twister'); %initialize Random Generator to make results replicable
gen_output_min = 0;   %minimal output of generators
gen_output_max = 400; %maximum output of generators
number_scenarios = 1000;

P_inj_test = (gen_output_max - gen_output_min) .* rand(length(mpc.bus(2:end,3)),number_scenarios) + gen_output_min; %generation of uniformly distribution generator outputs
%%
%preallocation of memory
%Error_sqr = zeros(number_scenarios,1);
Error_rel = zeros(number_scenarios,length(PTDF_R(:,1)));

for i=1:number_scenarios,
P_flow_zones_actual = mpc.zoneBranchSumID * PTDF * mpc.busWoS(:,3);
P_flow_zones_reduced = PTDF_R * gamma_R * mpc.busWoS(:,3);
Error_abs = abs(P_flow_zones_reduced-P_flow_zones_actual);
Error_rel = (P_flow_zones_reduced./P_flow_zones_actual -1) * 100;
%Error_sqr(i) = sqrt( 1/length(P_flow_zones_reduced) * sum((P_flow_zones_reduced - P_flow_zones_actual).^2));
end


%% Perform a multivariate linear regression to find out which bus injection variations causes errors
P_inj_test_rel = (bsxfun (@rdivide, P_inj_test', mpc.bus(2:end,3)')-1) .* 100; %Relative change of bus injections (related to base case)
%%
%Reg_node2branch
reg_node2branch = regstats(Error_rel(:,1), P_inj_test_rel); 

%% In-depth analysis of the impact of variations of bus injections

%simulation parameter
steps = 100;
var = 50; %in MW
%gen_out_max = base2 + 50;
%gen_out_min = base2 - 50; 
	
%memory preallocation
P_inj_test_bus = kron(ones(1,steps),mpc.bus(2:end,3));
Error_rel_bus = cell(length(mpc.bus(2:end,1)),1);
Error_rel = zeros(steps,length(P_flow_zones_reduced));

%error calculation of bus j
for j=1:length(mpc.bus(2:end,1)),
P_inj_test_bus = kron(ones(1,steps),mpc.bus(2:end,3));
%P_inj_test_bus(j,:) = gen_output_min:gen_output_max/steps:gen_output_max;
P_inj_test_bus(j,:) = mpc.bus(j+1,3)-(var-1):2*var/steps:mpc.bus(j+1,3)+var;

    for i=1:steps
    P_flow_zones_actual = mpc.zoneBranchSumID * PTDF * P_inj_test_bus(:,i);
    P_flow_zones_reduced = PTDF_R * gamma_R * P_inj_test_bus(:,i);
    Error_rel(i,:) = (P_flow_zones_reduced./P_flow_zones_actual -1) * 100;
    end
    Error_rel_bus{j} = Error_rel;
end
%%
node_all = zeros(length(5*length(mpc.branchR(:,1))),4);
for i=1:length(mpc.bus(2:end,1))
%node = error_calc(i,Error_rel_bus); 
% node_all(1+(i-1)*(length(mpc.branchR(:,1))+1):length(mpc.branchR(:,1))+(i-1)*(length(mpc.branchR(:,1))+1),1:4) = error_calc(i,Error_rel_bus);
end
%xlswrite('error_evaluation.xlsx',node,'node_variation',strcat('B',num2str(2+6*(i-1)),':E',num2str(6+6*(i-1))));
xlswrite('error_evaluation.xlsx',node_all,'node_variation','B2:E30'); %loops are very slow therefore the matrix has to be written at once
